/*
 * i2sbus driver
 *
 * Copyright 2006 Johannes Berg <johannes@sipsolutions.net>
 *
 * GPL v2, can be found in COPYING.
 */

#include <linux/module.h>
#include <asm/macio.h>
#include <asm/dbdma.h>
#include <linux/pci.h>
#include <linux/interrupt.h>
#include <asm/pmac_feature.h>
#include <asm/pmac_pfunc.h>
#include <sound/driver.h>
#include <sound/core.h>
/* FIX THIS */
#include "../soundbus.h"
#include "i2sbus.h"
/* should go away, see comment in i2sbus_add_dev */
#include <linux/dma-mapping.h>

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Johannes Berg <johannes@sipsolutions.net>");
MODULE_DESCRIPTION("Apple Soundbus: I2S support");
/* for auto-loading, declare that we handle this weird
 * string that macio puts into the relevant device */
MODULE_ALIAS("of:Ni2sTi2sC");

static struct of_device_id i2sbus_match[] = {
	{ .name = "i2s" },
	{ }
};

static int alloc_dbdma_descriptor_ring(struct i2sbus_dev *i2sdev,
				       struct dbdma_command_mem *r, int numcmds)
{
	/* one more for rounding */
	r->size = (numcmds+1) * sizeof(struct dbdma_cmd);
        /* We use the PCI APIs for now until the generic one gets fixed
         * enough or until we get some macio-specific versions
         */
	r->space = pci_alloc_consistent(macio_get_pci_dev(i2sdev->macio),
					r->size,
					&r->bus_addr);

	if (!r->space) return -ENOMEM;

	memset(r->space, 0, r->size);
	r->cmds = (void*)DBDMA_ALIGN(r->space);
	r->bus_cmd_start = r->bus_addr + (dma_addr_t)((char*)r->cmds - (char*)r->space);

	return 0;
}

static void free_dbdma_descriptor_ring(struct i2sbus_dev *i2sdev,
				      struct dbdma_command_mem *r)
{
	if (!r->space) return;
	
	pci_free_consistent(macio_get_pci_dev(i2sdev->macio), r->size, r->space, r->bus_addr);
}

static void i2sbus_release_dev(struct device *dev)
{
	struct i2sbus_dev *i2sdev = container_of(dev, struct i2sbus_dev, sound.ofdev.dev);
	int i;

	printk(KERN_INFO "freeing i2sbus device\n");

 	if (i2sdev->intfregs) iounmap(i2sdev->intfregs);
 	if (i2sdev->out.dbdma) iounmap(i2sdev->out.dbdma);
 	if (i2sdev->in.dbdma) iounmap(i2sdev->in.dbdma);
	for (i=0;i<3;i++)
		if (i2sdev->resource_allocated & (1<<i))
			release_mem_region(i2sdev->resources[i].start,
					   i2sdev->resources[i].end -
					   i2sdev->resources[i].start + 1);
	free_dbdma_descriptor_ring(i2sdev, &i2sdev->out.dbdma_ring);
	free_dbdma_descriptor_ring(i2sdev, &i2sdev->in.dbdma_ring);
	for (i=0;i<3;i++)
		free_irq(i2sdev->interrupts[i], i2sdev);
	i2sbus_control_release(i2sdev->control);
	kfree(i2sdev);
}

static irqreturn_t i2sbus_bus_intr(int irq, void *devid, struct pt_regs *regs)
{
	struct i2sbus_dev *dev = devid;
	u32 intreg = in_le32(&dev->intfregs->intr_ctl);

	printk(KERN_INFO "i2sbus: interrupt, intr reg is 0x%x!\n", intreg);

	/* acknowledge interrupt reasons */
	out_le32(&dev->intfregs->intr_ctl, intreg);

	return IRQ_HANDLED;
}

static int i2sbus_add_dev(struct macio_dev *macio, struct i2sbus_control *control,
			  struct device_node *np)
{
	struct i2sbus_dev *dev;
	struct device_node *child = NULL, *sound = NULL;
	int i;
	static const char *rnames[] = { "i2sbus control", "i2sbus tx", "i2sbus rx" };
	static irqreturn_t (*ints[])(int irq, void *devid, struct pt_regs *regs) = {
		i2sbus_bus_intr,
		i2sbus_tx_intr,
		i2sbus_rx_intr
	};
	struct pmf_args args = { .count = 0 };

	if (np->n_intrs != 3)
		return 0;

	dev = kzalloc(sizeof(struct i2sbus_dev), GFP_KERNEL);
	if (!dev)
		return 0;

	i = 0;
	while ((child = of_get_next_child(np, child))) {
		if (strcmp(child->name, "sound") == 0) {
			i++;
			sound = child;
		}
	}
	if (i == 1) {
		u32 *layout_id;
		layout_id = (u32*) get_property(sound, "layout-id", NULL);
		if (layout_id) {
			snprintf(dev->sound.modalias, 32, "sound-layout-%d", *layout_id);
		}
	}

	dev->sound.ofdev.node = np;
	dev->sound.ofdev.dma_mask = macio->ofdev.dma_mask;
	dev->sound.ofdev.dev.dma_mask = &dev->sound.ofdev.dma_mask;
	dev->sound.ofdev.dev.parent = &macio->ofdev.dev;
	dev->sound.ofdev.dev.release = i2sbus_release_dev;
	dev->sound.attach_codec = i2sbus_attach_codec;
	dev->sound.detach_codec = i2sbus_detach_codec;
	dev->sound.pcmid = -1;
	dev->macio = macio;
	dev->control = control;
	macio->ofdev.dev.driver_data = dev;
	INIT_LIST_HEAD(&dev->sound.codec_list);

	for (i=0;i<3;i++) {
		if (request_irq(np->intrs[i].line, ints[i], 0, rnames[i], dev))
			goto err;
		dev->interrupts[i] = np->intrs[i].line;
	}

	/* FIXME (need help from BenH):
	 * We should be using the macio device to get acess
	 * to our resources IMHO, but since we are not actually
	 * the macio-device itself (since we're a child of that),
	 * we currently can't.
	 */
	for (i=0;i<3;i++) {
		if (of_address_to_resource(np, i, &dev->resources[i]))
			goto err;
		if (request_mem_region(dev->resources[i].start,
				       dev->resources[i].end -
				       dev->resources[i].start + 1,
				       rnames[i]) == NULL)
			goto err;
		dev->resource_allocated |= (1<<i);
	}
	/* should do sanity checking here about length of them */
	dev->intfregs = ioremap(dev->resources[0].start,
			       dev->resources[0].end-dev->resources[0].start+1);
	dev->out.dbdma = ioremap(dev->resources[1].start,
				dev->resources[1].end-dev->resources[1].start+1);
	dev->in.dbdma = ioremap(dev->resources[2].start,
				dev->resources[2].end-dev->resources[2].start+1);
	if (!dev->intfregs || !dev->out.dbdma || !dev->in.dbdma)
		goto err;

	printk(KERN_INFO "i2sbus: found i2s controller\n");

	if (alloc_dbdma_descriptor_ring(dev, &dev->out.dbdma_ring, MAX_DBDMA_COMMANDS))
		goto err;
	if (alloc_dbdma_descriptor_ring(dev, &dev->in.dbdma_ring, MAX_DBDMA_COMMANDS))
		goto err;

	if (soundbus_add_one(&dev->sound)) {
		printk(KERN_DEBUG "i2sbus: device registration error!\n");
		goto err;
	}

	/* this belongs to the control layer */
	pmf_call_function(np, "cell-enable", &args);
	pmf_call_function(np, "enable", &args);
	pmf_call_function(np, "clock-enable", &args);

	/* print out bus regs */
	printk("serial format: 0x%x\n", ld_le32(&dev->intfregs->serial_format));
	printk("dws: 0x%x\n", ld_le32(&dev->intfregs->data_word_sizes));

	return 1;
 err:
	for (i=0;i<3;i++)
		if (dev->interrupts[i])
			free_irq(dev->interrupts[i], dev);
	free_dbdma_descriptor_ring(dev, &dev->out.dbdma_ring);
	free_dbdma_descriptor_ring(dev, &dev->in.dbdma_ring);
	if (dev->intfregs) iounmap(dev->intfregs);
	if (dev->out.dbdma) iounmap(dev->out.dbdma);
	if (dev->in.dbdma) iounmap(dev->in.dbdma);
	for (i=0;i<3;i++)
		if (dev->resource_allocated & (1<<i))
			release_mem_region(dev->resources[i].start,
					   dev->resources[i].end -
					   dev->resources[i].start + 1);
	kfree(dev);
	return 0;
}

static int i2sbus_probe(struct macio_dev* dev, const struct of_device_id *match)
{
	struct device_node *np = NULL;
	int got = 0, err;
	struct i2sbus_control *control = NULL;

	err = i2sbus_control_init(dev, &control);
	if (err)
		return err;
	if (!control) {
		printk(KERN_ERR "i2sbus_control_init API breakage\n");
		return -ENODEV;
	}

	while ((np = of_get_next_child(dev->ofdev.node, np))) {
		/* FIXME: add "i2s-modem" here, but only after
		 * we get rid of the platform function foo */
		if (device_is_compatible(np, "i2sbus")) {
			got += i2sbus_add_dev(dev, control, np);
		}
	}

	control->usecount = got;
	if (got)
		return 0;

	/* found none, clean up */
	i2sbus_control_destroy(control);
	return -ENODEV;
}

static int i2sbus_remove(struct macio_dev* dev)
{
	struct i2sbus_dev* i2sdev = (struct i2sbus_dev*)dev->ofdev.dev.driver_data;
	soundbus_remove_one(&i2sdev->sound);

	return 0;
}

#ifdef CONFIG_PM
static int i2sbus_suspend(struct macio_dev* dev, pm_message_t state)
{
	struct codec_info_item *cii;
	struct i2sbus_dev* i2sdev = (struct i2sbus_dev*)dev->ofdev.dev.driver_data;
	int err, ret = 0;

	printk("i2sbus_suspend()\n");

	/* Notify Alsa */
	if (i2sdev->sound.pcm) {
		/* Suspend PCM streams */
		snd_pcm_suspend_all(i2sdev->sound.pcm);
		/* Probably useless as we handle power transitions ourselves */
		snd_power_change_state(i2sdev->sound.pcm->card, SNDRV_CTL_POWER_D3hot);
	}
	/* Notify codecs */
	list_for_each_entry(cii, &i2sdev->sound.codec_list, list) {
		err = 0;
		if (cii->codec->suspend) err = cii->codec->suspend(cii, state);
		if (err) ret = err;
	}
	return ret;
}

static int i2sbus_resume(struct macio_dev* dev)
{
	struct codec_info_item *cii;
	struct i2sbus_dev* i2sdev = (struct i2sbus_dev*)dev->ofdev.dev.driver_data;
	int err, ret = 0;

	printk("i2sbus_resume()\n");

	/* Notify codecs so they can re-initialize */
	list_for_each_entry(cii, &i2sdev->sound.codec_list, list) {
		err = 0;
		if (cii->codec->resume) err = cii->codec->resume(cii);
		if (err) ret = err;
	}
	/* Notify Alsa */
        if (i2sdev->sound.pcm) {
		/* Same comment as above, probably useless */
		snd_power_change_state(i2sdev->sound.pcm->card, SNDRV_CTL_POWER_D0);
	}

	return ret;
}
#endif /* CONFIG_PM */

static int i2sbus_shutdown(struct macio_dev* dev)
{
	return 0;
}

static struct macio_driver i2sbus_drv = {
	.name = "soundbus-i2s",
	.owner = THIS_MODULE,
	.match_table = i2sbus_match,
	.probe = i2sbus_probe,
	.remove = i2sbus_remove,
#ifdef CONFIG_PM
	.suspend = i2sbus_suspend,
	.resume = i2sbus_resume,
#endif
	.shutdown = i2sbus_shutdown,
};

static int __init soundbus_i2sbus_init(void)
{
	return macio_register_driver(&i2sbus_drv);
}

static void __exit soundbus_i2sbus_exit(void)
{
	macio_unregister_driver(&i2sbus_drv);
}

module_init(soundbus_i2sbus_init);
module_exit(soundbus_i2sbus_exit);
